/*////////////////////////////////////////////////////////////////////////
   Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

   Permission to use, copy, modify, and distribute this material for any
   purpose and without fee is hereby granted, provided that the above
   copyright notice and this permission notice appear in all copies, and
   that the name of ETL not be used in advertising or publicity pertaining
   to this material without the specific, prior written permission of an
   authorized representative of ETL.
   ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
   MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
   OR IMPLIED WARRANTIES.
   /////////////////////////////////////////////////////////////////////////
   Content-Type:        program/C; charset=US-ASCII
   Program:     ciiserv.c (CII font server/client)
   Author:              Yutaka Sato <ysato@etl.go.jp>
   Description:
   History:
   951001       created
   /////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <string.h>
#include "cii.h"
extern FILE *Tmpfile ();
extern char *strdup ();
#define VERSION	"0.2"
char *CII_version = VERSION;
#define only16H(ciiver)	(strcmp(ciiver,"0.2") < 0)

#define A_7bit	"7bit"
#define A_8bit	"8bit"
#define A_1byte	"1byte"
#define A_2byte	"2byte"

#define XBMSIZE(w,h)	((w + 7) / 8 * h)
#define DELIMITER "--\r\n"

extern struct CharCodesXBM CII_CC[];
extern int CII_CC_SIZE;

typedef struct {
	unsigned char *bmap;	/* bitmap for xbmind */
	int bmapsize;		/* num. of bytes of bmap */
	FILE *xbmfp;		/* the file contains xbmsrc */
	int xbmoff;		/* start position in xbmfp */
	int xbmtotal;		/* total num. of bytes of xbmsrc */
	int last_code;		/* cache for XBMIND() */
	int last_xbmi;		/* cache for XBMIND() */
} XbmCache;

#define EXTP		((XbmCache*)csx->ext)
#define EXT(elem)	(EXTP ? EXTP->elem : 0)
#define BMAP		EXT(bmap)
#define BMAPSIZE	EXT(bmapsize)
#define XBMFP		EXT(xbmfp)
#define XBMOFF		EXT(xbmoff)
#define XBMTOTAL	EXT(xbmtotal)
#define LASTCODE	EXT(last_code)
#define LASTXBMI	EXT(last_xbmi)

static
bmapcnt (csx, code)
	struct CharCodesXBM *csx;
{
	int nb, xi, xj, map, mask, xbmi;

	if (EXTP == NULL)
		return -1;

	if (code == LASTCODE)
		return LASTXBMI;

	nb = 0;
	xbmi = -1;
	for (xi = 0; xi <= code; xi += 8) {
		map = BMAP[xi / 8];
		mask = 1;
		for (xj = 0; xj < 8; xj++) {
			if (xi + xj == code) {
				xbmi = map & mask ? nb : -1;
				break;
			}
			if (map & mask)
				nb++;
			mask = (mask << 1);
		}
	}
	EXTP->last_code = code;
	EXTP->last_xbmi = xbmi;
	return xbmi;
}

XBMIND (csx, idx)
	struct CharCodesXBM *csx;
{
	if (csx->first - csx->last < idx)
		return -1;
	if (csx->xbmind)
		return csx->xbmind[idx];
	if (BMAP)
		return bmapcnt (csx, idx);
	return -1;
}

char *
XBMSRC (csx, idx, buf)
	struct CharCodesXBM *csx;
	char *buf;
{
	int off, bsize;

	if (csx->xbmsrc)
		return csx->xbmsrc[XBMIND (csx, idx)];
	else {
		bsize = XBMSIZE (csx->width, csx->height);
		off = XBMOFF + XBMIND (csx, idx) * bsize;
		fseek (XBMFP, off, 0);
		fread (buf, 1, bsize, XBMFP);
		return buf;
	}
}

CII_lsFonts (buff, size)
	char *buff;
{
	int ci;
	char *bp;
	struct CharCodesXBM *csx;

	bp = buff;
	for (ci = 0; ci < CII_CC_SIZE; ci++) {
		csx = &CII_CC[ci];
		if (csx->name == NULL)
			break;
		sprintf (bp, "%-10s %s/%dx%d\n", csx->name, csx->regname,
			 csx->width, csx->height);
		bp += strlen (bp);
	}
	return ci;
}

static
scanCharSetXBM (name, create, matches)
	char *name;
	struct CharCodesXBM *matches[];
{
	int ci;
	struct CharCodesXBM *csx;
	char fname[128];
	int width, height;
	int nmatch;

	fname[0] = 0;
	width = height = 0;
	sscanf (name, "%[^/]/%dx%d", fname, &width, &height);

	nmatch = 0;
	for (ci = 0; ci < CII_CC_SIZE; ci++) {
		matches[ci] = 0;
		csx = &CII_CC[ci];
		if (csx->name == NULL)
			break;
		if (strcmp (fname, "*") == 0
		    || strcasecmp (fname, CII_CC[ci].name) == 0
		    || strcasecmp (fname, CII_CC[ci].regname) == 0) {
			if (width == 0 || width == csx->width)
				if (height == 0 || height == csx->height)
					matches[nmatch++] = csx;
		}
	}
	if (create && nmatch == 0 && ci < CII_CC_SIZE)
		matches[nmatch++] = csx;
	return nmatch;
}

static struct CharCodesXBM *
CharSetXBM (name, create)
	char *name;
{
	struct CharCodesXBM *matches[128];
	int nmatch;

	nmatch = scanCharSetXBM (name, create, matches);
	if (nmatch)
		return matches[0];
	else
		return NULL;
}

CII_withCharset (charset)
	char *charset;
{
	return CharSetXBM (charset, 0) != NULL;
}

static char *
Sprintf (out, fmt, a, b, c, d, e, f, g)
	char *out, *fmt, *a, *b, *c, *d, *e, *f, *g;
{
	sprintf (out, fmt, a, b, c, d, e, f, g);
	return out + strlen (out);
}

CII_putFont1X (csx, tc)
	struct CharCodesXBM *csx;
	FILE *tc;
{
	int leng;
	char head[2048], *hp;
	int map, mapx;
	char *bp;
	int nx, xi, nb, bsize, bi;
	char xbmbuf[2048];

	hp = head;
	hp = Sprintf (hp, "CII-Version: %s\n", VERSION);
	hp = Sprintf (hp, "Name: %s,%s\n", csx->regname, csx->name);
	hp = Sprintf (hp, "Size: %d,%d\n", csx->width, csx->height);
	hp = Sprintf (hp, "Range: %d,%d\n", csx->first, csx->last);
	hp = Sprintf (hp, "Attrs: ");
	if (csx->is2bytes)
		hp = Sprintf (hp, A_2byte);
	else
		hp = Sprintf (hp, A_1byte);
	hp = Sprintf (hp, ",");
	if (csx->has8bit)
		hp = Sprintf (hp, A_8bit);
	else
		hp = Sprintf (hp, A_7bit);
	hp = Sprintf (hp, "\n");
	hp = Sprintf (hp, "\f\n\n");
	fputs (head, tc);
	leng = strlen (head);

	if (XBMFP) {
		fwrite (BMAP, 1, BMAPSIZE, tc);
		leng += BMAPSIZE;
		fseek (XBMFP, XBMOFF, 0);
		leng += filecopy (tc, XBMFP, XBMTOTAL);
		return leng;
	}

	nx = csx->last - csx->first + 1;
	map = 0;
	mapx = 0;
	nb = 0;
	for (xi = 0; xi < nx; xi++) {
		if (0 <= XBMIND (csx, xi)) {
			nb++;
			map |= (1 << mapx);
		}
		if (++mapx == 8) {
			putc (map, tc);
			leng++;
			map = 0;
			mapx = 0;
		}
	}
	if (mapx != 0) {
		putc (map, tc);
		leng++;
	}

	bsize = XBMSIZE (csx->width, csx->height);
	nb = 0;
	for (xi = 0; xi < nx; xi++) {
		if (XBMIND (csx, xi) < 0)
			continue;
		bp = XBMSRC (csx, xi, xbmbuf);
		for (bi = 0; bi < bsize; bi++)
			putc (bp[bi], tc);
		nb++;
	}
	leng += nb * bsize;
	return leng;
}

CII_putFont1 (charset, tc)
	char *charset;
	FILE *tc;
{
	struct CharCodesXBM *csx;

	if (csx = CharSetXBM (charset, 0))
		return CII_putFont1X (csx, tc);
	else
		return 0;
}

static
filecopy (dst, src, len)
	FILE *dst, *src;
	int len;
{
	int rem, rc1, rc2;
	char buf[2048];

	rem = len;
	while (0 < rem) {
		rc1 = sizeof (buf) < rem ? sizeof (buf) : rem;
		rc2 = fread (buf, 1, rc1, src);
		fwrite (buf, 1, rc2, dst);
		rem -= rc2;
	}
	return len - rem;
}

CII_getFont1 (charset, fs, fp)
	char *charset;
	FILE *fs, *fp;
{
	struct CharCodesXBM *csx, csx0;
	char head1[1024], name[1024], value[1024];
	char rname[1024], alias[1024], frname[1024];
	char val1[256], val2[256];
	int leng;
	int nx, xi, map, mapx, cx, bytes, nb, bsize, rcc;
	char *xbmsrc, xbmbuf[1024];

	csx = &csx0;
	bzero (&csx0, sizeof (csx0));

	leng = 0;
	rname[0] = alias[0] = 0;
	while (fgets (head1, sizeof (head1), fs) != NULL) {
		leng += strlen (head1);
		if (head1[0] == '\r' || head1[0] == '\n')
			break;
		if (sscanf (head1, "%[^:]: %s", name, value) != 2)
			continue;
		val1[0] = val2[0] = 0;
		sscanf (value, "%[^,],%s", val1, val2);

		if (strcasecmp (name, "CII-Version") == 0) {
			if (strcmp (value, VERSION) != 0) {
				CII_debug ("Warning: CII version mismatch[%s : %s]\n",
					   value, VERSION);
			}
		}
		else if (strcasecmp (name, "Name") == 0) {
			strcpy (alias, val2);
			strcpy (rname, val1);
		}
		else if (strcasecmp (name, "Size") == 0) {
			csx->width = atoi (val1);
			csx->height = atoi (val2);
		}
		else if (strcasecmp (name, "Range") == 0) {
			csx->first = atoi (val1);
			csx->last = atoi (val2);
		}
		else if (strcasecmp (name, "Attrs") == 0) {
			csx->is2bytes = strstr (value, A_2byte) != 0;
			csx->has8bit = strstr (value, A_8bit) != 0;
		}
	}
	if (rname[0] == 0 || alias[0] == 0)
		return leng;

	sprintf (frname, "%s/%dx%d", rname, csx->width, csx->height);
	csx = CharSetXBM (frname, 1);
	*csx = csx0;

	csx->name = strdup (alias);
	csx->regname = strdup (rname);

	nx = csx->last - csx->first + 1;
	nb = 0;

	csx->ext = (char *) calloc (1, sizeof (XbmCache));
	EXTP->bmapsize = (nx + 7) / 8;
	EXTP->bmap = (unsigned char *) malloc (BMAPSIZE);
	for (xi = 0; xi < nx; xi += 8) {
		map = getc (fs);
		EXTP->bmap[xi / 8] = map;
		leng++;
		for (mapx = 0; mapx < 8 && xi + mapx < nx; mapx++)
			if (map & (1 << mapx))
				nb++;
	}

	bsize = XBMSIZE (csx->width, csx->height);

	if (fp != NULL) {
		EXTP->xbmfp = fp;
		EXTP->xbmoff = ftell (fp);
		rcc = filecopy (fp, fs, nb * bsize);
		EXTP->xbmtotal = rcc;
		leng += rcc;
	}
	else {
		xbmsrc = (char *) malloc (nb * bsize);
		rcc = fread (xbmsrc, 1, nb * bsize, fs);
		leng += rcc;

		bytes = XBMSIZE (csx->width, csx->height);
		csx->xbmsrc = (char **) malloc (sizeof (char *) * nb);
		for (xi = 0; xi < nb; xi++)
			csx->xbmsrc[xi] = &xbmsrc[xi * bytes];
	}

	return leng;
}

CII_getFonts (charset, fs)
	char *charset;
	FILE *fs;
{
	int leng = 0;
	char delimiter[8];
	FILE *fp;

	fp = Tmpfile ();
	for (;;) {
		leng += CII_getFont1 (charset, fs, fp);
		if (feof (fs))
			break;
		fgets (delimiter, sizeof (delimiter), fs);
		leng += strlen (delimiter);
		if (strcmp (delimiter, DELIMITER) != 0)
			break;
	}
	if (fp != NULL)
		fflush (fp);
	return leng;
}

static
isin (size, csxv, csx)
	struct CharCodesXBM *csxv[], *csx;
{
	int cx;
	for (cx = 0; cx < size; cx++)
		if (csxv[cx] == csx)
			return 1;
	return 0;
}

static
bundleFamilyFonts (nfont, fontv)
	struct CharCodesXBM *fontv[];
{
	int putall1bytes = 0;
	int cx, ci;
	struct CharCodesXBM *csx;

	for (cx = 0; cx < nfont; cx++) {
		csx = fontv[cx];
		if (!csx->is2bytes)
			putall1bytes = 1;
	}

	if (putall1bytes)
		for (ci = 0; ci < CII_CC_SIZE; ci++) {
			csx = &CII_CC[ci];
			if (csx->name == NULL)
				break;
			if (!csx->is2bytes && !isin (nfont, fontv, csx))
				fontv[nfont++] = csx;
		}

	return nfont;
}

CII_putFontsX (fs, nfont, fontv, ciiver)
	FILE *fs;
	struct CharCodesXBM *fontv[];
	char *ciiver;
{
	int leng = 0, leng1;
	int cx;
	struct CharCodesXBM *csx;

	for (cx = 0; cx < nfont; cx++) {
		csx = fontv[cx];
		if (only16H (ciiver) && csx->height != 16)
			continue;

		if (cx != 0) {
			fputs (DELIMITER, fs);
			leng += strlen (DELIMITER);
		}
		leng += leng1 = CII_putFont1X (csx, fs);
		CII_debug ("[%2d/%2d] %s/%dx%d: %d (%s)\n", cx + 1, nfont,
		   csx->regname, csx->width, csx->height, leng1, csx->name);
	}
	CII_debug ("total: %d\n", leng);
	return leng;
}

CII_putFonts (charset, ciiver, fs)
	char *charset, *ciiver;
	FILE *fs;
{
	struct CharCodesXBM *fontv[128];
	int nfont;

	if (ciiver == NULL)
		ciiver = "0.1";
	CII_debug ("ClientVersion: %s\n", ciiver);
	nfont = scanCharSetXBM (charset, 0, fontv);
	if (nfont == 0)
		return 0;
	nfont = bundleFamilyFonts (nfont, fontv);
	return CII_putFontsX (fs, nfont, fontv, ciiver);
}

int (*CII_debugprinter) ();
CII_debug (fmt, a, b, c, d, e, f, g)
	char *fmt, *a, *b, *c, *d, *e, *f, *g;
{
	if (CII_debugprinter)
		(*CII_debugprinter) (fmt, a, b, c, d, e, f, g);
}


#ifdef MAIN
static
fprintf_stderr (fmt, a, b, c, d, e, f, g)
{
	return fprintf (stderr, fmt, a, b, c, d, e, f, g);
}
main (ac, av)
	char *av[];
{
	int ai, leng;
	char *charset;
	FILE *out = stdout;

	CII_debugprinter = fprintf_stderr;

	if (ac <= 1) {
		CII_debug ("Usage: %s font1 font2 ...\n", av[0]);
		exit (1);
	}
	for (ai = 1; ai < ac; ai++) {
		charset = av[ai];
		leng = CII_putFonts (charset, out);
		if (leng <= 0) {
			if (1 < ac)
				fputs (DELIMITER, out);
			CII_debug ("%s: not found.\n", charset);
			exit (1);
		}
	}
	exit (0);
}
#endif
