/************************************************************************
* Copyright (c) 1995 Valentin Shopov, Blagoevgrad, Bulgaria             *
*                                                                       *
* Permission to use, copy, modify, and distribute this material for any *
* purpose and without fee is hereby granted, provided that the above    *
* copyright notice and this permission notice appear in all copies, and *
* that the name of the author not be used in advertising or publicity   *
* pertaining to this material without the specific, prior written       *
* permission of the author.                                             *
*                                                                       *
* THE AUTHOR MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY *
* OF THIS MATERIAL FOR ANY PURPOSE. IT IS PROVIDED "AS IS", WITHOUT ANY *
* EXPRESS OR IMPLIED WARRANTIES.                                        *
*************************************************************************/

/**************************ciilib2.1.7***********************************
 *                                                                      *
 *        Character to Inline Image Conversion Library (CIILIB)         *
 *         This version of CIILIB with the following functions 		*
 *			by Mr. Akira Kitauchi				*
 *	improves availability and speed:				*
 *       - serves CII with the image/GIF content type			*
 *       - serves CII word by word (not character by character)		*
 *									*
 *                                                                      *
 *    GIF coded CII							*
 *              August 1995, by Akira KITAUCHI				*
 *									*
 *              E-mail address: kitauchi@i.h.kyoto-u.ac.jp		*
 *              WWW home page : http://www.i.h.kyoto-u.ac.jp/~kitauchi/	*
 ************************************************************************/

#include <stdio.h>
#include "cii.h"
extern char *XBMSRC ();

static int isXBM;
static char *xbmname;

/*----
 *	text to CII/GIF converter
 */

/*----
 *	CII/GIF font server
 */
static void
expand_H1 (srcb, dstb, width)
	char *srcb, *dstb;
	int width;
{
	int expb, expb0, dmask, smask, i, j;

	expb0 = 0;

	for (i = 0; i < width; i++) {
		expb = expb0;
		dmask = 3;
		smask = 1;
		for (j = 0; j < 8; j++) {
			expb |= (srcb[i] & smask) ? dmask : 0;
			smask <<= 1;
			dmask <<= 2;
		}
		dstb[i * 2 + 0] = (char) expb;
		dstb[i * 2 + 1] = (char) (expb >> 8);
		expb0 = expb >> 16;
	}
}

static void
shrink_PIC (expbuf, picbuf, chN, chW, chH)
	char *expbuf, *picbuf;
	int chN, chW, chH;
{
	int h, n, w, isBRK;
	memset (picbuf, 0, ((chN * chW + 7) / 8) * chH);
	for (h = 0; h < chH; h++) {
		for (n = 0, isBRK = 0;; n++) {
			for (w = 0; w < 8; w++) {
				if (8 * n + w + 1 > chN * chW) {
					isBRK = (w == 0) ? 2 : 1;
					break;
				}
				*picbuf |= (*expbuf << w);
				expbuf++;
			}
			picbuf++;
			if (isBRK) {
				picbuf -= (isBRK - 1);
				break;
			}
		}
	}
}

extern void expand_PIC ();
static int
PIC_to_XBM (picbuf, xbmbuf, chN, chW, chH, color, isH1)
	char *picbuf, *xbmbuf;
	int chN, chW, chH;
	int color, isH1;
{
	int i;
	char *xbm;
	char expbuf[32 * 34 * (GIFWIDTH + 1)];
	if ((chW % 8) != 0 && chN > 1) {
		expand_PIC (picbuf, expbuf, chN, chW, chH, isH1);
		shrink_PIC (expbuf, picbuf, chN, chW, chH);
	}

	xbm = xbmbuf;
	sprintf (xbm, "#define %s_width %d\n#define %s_height %d\nstatic char %s_bits[] = {\0", xbmname, chW * chN, xbmname, chH, xbmname);
	for (i = 0; i < (chN * chW + 7) / 8 * chH; i++) {
		if (i % 16 == 0)
			sprintf (xbm + strlen (xbm), "\n\0");
		sprintf (xbm + strlen (xbm), "0x%02x,\0", (unsigned char) picbuf[i]);
	}
	sprintf (xbm + strlen (xbm) - 1, "};\n\0");
	return (int) strlen (xbm);
}

extern int PIC_to_GIF ();
static int
PIC_to_IMG (picbuf, gifbuf, chN, chW, chH, color, isH1)
	char *picbuf, *gifbuf;
	int chN, chW, chH;
	int color;
{
	if (isXBM)
		return PIC_to_XBM (picbuf, gifbuf, chN, chW, chH, color, isH1);
	else
		return PIC_to_GIF (picbuf, gifbuf, chN, chW, chH, color, isH1);
}

static int
getCII_H1 (picbuf, gifbuf, color, chN, chW, chH, isU)
	char *picbuf, *gifbuf;
	int color, chN, chW, chH, isU;
{
	char *picbuf2;
	int i, j, w, rc;

	w = ((chW + 7) / 8) * chN;
	picbuf2 = (char *) malloc (w * 2 * (chH * 2 + (isU ? 2 : 0)));

	for (i = 0; i < chH; i++) {
		expand_H1 (picbuf + i * w, picbuf2 + i * 2 * w * 2, w);
		memcpy (picbuf2 + (i * 2 + 1) * w * 2, picbuf2 + i * 2 * w * 2, w * 2);
	}

	chH *= 2;
	if (isU) {
		memset (picbuf2 + w * 2 * (chH + 0), 0x00, w * 2);
		memset (picbuf2 + w * 2 * (chH + 1), 0xff, w * 2);
		chH += 2;
	}

	rc = PIC_to_GIF (picbuf2, gifbuf, chN * 2, chW, chH, color, 1);
	free (picbuf2);
	return rc;
}


static int
getCII (picbuf, gifbuf, color, chN, chW, chH, isU)
	char *picbuf, *gifbuf;
	int color, chN, chW, chH, isU;
{
	int w;

	if (isU) {
		w = ((chW + 7) / 8) * chN;
		memset (picbuf + w * (chH + 0), 0x00, w);
		memset (picbuf + w * (chH + 1), 0xff, w);
		chH += 2;
	}

	return PIC_to_IMG (picbuf, gifbuf, chN, chW, chH, color, 0);
}

static int
getCII_BOLD (picbuf, gifbuf, color, chN, chW, chH, isU)
	char *picbuf, *gifbuf;
	int color, chN, chW, chH, isU;
{
	int i, j, bit0, bit7, w, len;
	unsigned char *pb;

	pb = (unsigned char *) picbuf;
	w = ((chW + 7) / 8) * chN;

	for (i = 0; i < chH; i++) {
		bit0 = 0;
		for (j = 0; j < w; j++) {
			bit7 = *pb & 0x80;
			*pb |= (*pb << 1) | bit0;
			bit0 = bit7 >> 7;
			pb++;
		}
	}
	if (isU) {
		memset (picbuf + w * (chH + 0), 0x00, w);
		memset (picbuf + w * (chH + 1), 0xff, w);
		chH += 2;
	}

	return PIC_to_IMG (picbuf, gifbuf, chN, chW, chH, color, 0);
}

CII_getGIF (name, gifbuf)
	char *name, *gifbuf;
{
	extern struct CharCodesXBM CII_CC[];
	extern GIFCII_setccid ();
	char xbmbuf[1024];

	char charset[32], codebuf[GIFWIDTH * 4 + 32], *code, *picbuf;
	static char uspcbuf[50];
	int chW, chH, chN, chWbytes, Hsize;
	int color, isU = 0, isLAP = 0, rc, cc_id, len, picsize, i, j, k;
	struct CharCodesXBM *cc;

	code = codebuf;
	xbmname = (char *) strrchr (name, '/') + 1;

	sscanf (name, "%[^/]/%6x%s", charset, &color, code);

	code++;			/* ignore 'page type' */

	if (*code == 'x') {
		if (!color)
			isXBM = 1;
		else
			isXBM = 0;
		code++;
	}
	else
		isXBM = 0;

	Hsize = *code++;
	cc_id = GIFCII_setccid (Hsize, charset);
	if (cc_id == -1)
		return 0;
	cc = &CII_CC[cc_id];

	if (*code == 'u') {
		isU = 1;
		code++;
	}
	if (*code == '+') {
		isLAP = 1;
		code++;
	}

	chW = cc->width;
	chH = cc->height;
	chWbytes = (chW + 7) / 8;

	len = strlen (code);
	if (len & 3)
		return 0;
	chN = isLAP ? 1 : (len >> 2);
	if (chN > GIFWIDTH)
		return 0;

	picsize = chWbytes * chN * (chH + (isU ? 2 : 0));
	picbuf = (char *) malloc (picsize);
	memset (picbuf, 0, picsize);

	for (i = 0; *code; code += 4) {
		unsigned int c;
		char *curgif;

		if (i == GIFWIDTH)
			return 0;

		if (!strncmp (code, "USPC", 4))
			curgif = uspcbuf;
		else {
			sscanf (code, "%4x", &c);
			if (!cc->has8bit)
				c &= 0x7F7F;
			if (c < cc->first || c > cc->last || XBMIND (cc, c - cc->first) == -1)
				curgif = uspcbuf;
			else
				curgif = XBMSRC (cc, c - cc->first, xbmbuf);
		}

		for (j = 0; j < chH; j++)
			for (k = 0; k < chWbytes; k++)
				picbuf[(chN * j + i) * chWbytes + k] |= curgif[j * chWbytes + k];

		if (!isLAP)
			i++;
	}

	switch (Hsize) {
	case 'h':
		if (chH >= 24)
			rc = getCII (picbuf, gifbuf, color, chN, chW, chH, isU);
		else
			rc = getCII_H1 (picbuf, gifbuf, color, chN, chW, chH, isU);
		break;
	case 'i':
		rc = getCII_BOLD (picbuf, gifbuf, color, chN, chW, chH, isU);
		break;
	case 'j':
		rc = getCII_BOLD (picbuf, gifbuf, color, chN, chW, chH, isU);
		break;
	case 'k':
		rc = getCII (picbuf, gifbuf, color, chN, chW, chH, isU);
		break;
	}

	free (picbuf);
	return rc;
}


CII_putHTTP_GIF (tc, vno, gkan)
	FILE *tc;
	char *gkan;
{
	int size;
	int leng;
	int chW;
	char gifbuf[(GIFWIDTH + 4) * (34 * 5 + 10) + (GIFWIDTH + 4) * 34 * 5 / 16 + 100];

	if ((size = CII_getGIF (gkan, gifbuf)) == 0)
		return 0;
	if (isXBM)
		fprintf (tc, "Content-Type: image/x-xbitmap\r\n");
	else
		fprintf (tc, "Content-Type: image/gif\r\n");
	fprintf (tc, "Content-Length: %d\r\n", size);
	fprintf (tc, "X-CII-Version: %s\r\n", CIIVER);
	fprintf (tc, "\r\n");
	leng += fwrite (gifbuf, 1, size, tc);
	return leng;
}
